import importlib
import logging

import gevent
import pytest
import requests

from prometheus_client import REGISTRY
from pyramid.config import Configurator
from pyramid.httpexceptions import HTTPNoContent

from baseplate import Baseplate
from baseplate.clients.requests import ExternalRequestsClient
from baseplate.clients.requests import InternalRequestsClient
from baseplate.frameworks.pyramid import BaseplateConfigurator
from baseplate.frameworks.pyramid import StaticTrustHandler
from baseplate.lib import config
from baseplate.observers.prometheus import PrometheusBaseplateObserver
from baseplate.server import make_listener
from baseplate.server.wsgi import make_server

from . import TestBaseplateObserver

logger = logging.getLogger(__name__)


@pytest.fixture
def gevent_socket():
    try:
        gevent.monkey.patch_socket()
        yield
    finally:
        import socket

        importlib.reload(socket)
        gevent.monkey.saved.clear()


@pytest.fixture
def http_server(gevent_socket):
    class HttpServer:
        def __init__(self, address):
            self.url = f"http://{address[0]}:{address[1]}/"
            self.requests = []

        def handle_request(self, request):
            self.requests.append(request)
            return HTTPNoContent()

    server_bind_endpoint = config.Endpoint("127.0.0.1:0")
    listener = make_listener(server_bind_endpoint)
    server_address = listener.getsockname()
    http_server = HttpServer(server_address)

    baseplate = Baseplate()
    trust_handler = StaticTrustHandler(trust_headers=True)
    baseplate_configurator = BaseplateConfigurator(baseplate, header_trust_handler=trust_handler)
    configurator = Configurator()
    configurator.include(baseplate_configurator.includeme)
    configurator.add_route("test_view", "/")
    configurator.add_view(http_server.handle_request, route_name="test_view", renderer="json")
    wsgi_app = configurator.make_wsgi_app()

    server = make_server({"stop_timeout": "1 millisecond"}, listener, wsgi_app)
    server_greenlet = gevent.spawn(server.serve_forever)
    try:
        yield http_server
    finally:
        server_greenlet.kill()


@pytest.mark.parametrize("client_cls", [InternalRequestsClient, ExternalRequestsClient])
@pytest.mark.parametrize("client_name", [None, "", "complex.client$name"])
@pytest.mark.parametrize("method", ["DELETE", "GET", "HEAD", "OPTIONS", "PATCH", "PUT", "POST"])
def test_client_makes_client_span(client_cls, client_name, method, http_server):
    baseplate = Baseplate(
        {"myclient.filter.ip_allowlist": "127.0.0.0/8", "myclient.filter.port_denylist": "0"}
    )
    if client_name is None:
        baseplate.configure_context({"myclient": client_cls()})
    else:
        baseplate.configure_context({"myclient": client_cls(client_name=client_name)})

    observer = TestBaseplateObserver()
    baseplate.register(observer)

    with baseplate.server_context("test") as context:
        fn = getattr(context.myclient, method.lower())
        response = fn(http_server.url)

    assert response.status_code == 204

    server_span_observer = observer.children[0]
    assert len(server_span_observer.children) == 1

    client_span_observer = server_span_observer.children[0]
    assert client_span_observer.span.name == "myclient.request"
    assert client_span_observer.on_start_called
    assert client_span_observer.on_finish_called
    assert client_span_observer.on_finish_exc_info is None
    assert client_span_observer.tags["http.url"] == http_server.url
    assert client_span_observer.tags["http.method"] == method
    assert client_span_observer.tags["http.status_code"] == 204
    assert client_span_observer.tags["http.slug"] == (
        client_name if client_name is not None else "myclient"
    )


@pytest.mark.parametrize("client_cls", [InternalRequestsClient, ExternalRequestsClient])
def test_connection_error(client_cls):
    baseplate = Baseplate(
        {"myclient.filter.ip_allowlist": "127.0.0.0/8", "myclient.filter.port_denylist": "0"}
    )
    baseplate.configure_context({"myclient": client_cls()})

    observer = TestBaseplateObserver()
    baseplate.register(observer)

    bogus_url = "http://localhost:1/"
    with pytest.raises(requests.exceptions.ConnectionError):
        with baseplate.server_context("test") as context:
            context.myclient.get(bogus_url)

    server_span_observer = observer.children[0]
    assert len(server_span_observer.children) == 1

    client_span_observer = server_span_observer.children[0]
    assert client_span_observer.span.name == "myclient.request"
    assert client_span_observer.on_start_called
    assert client_span_observer.on_finish_called
    assert client_span_observer.on_finish_exc_info is not None
    assert client_span_observer.tags["http.url"] == bogus_url
    assert client_span_observer.tags["http.method"] == "GET"
    assert "http.status_code" not in client_span_observer.tags


def test_internal_client_sends_headers(http_server):
    baseplate = Baseplate()
    baseplate.configure_context({"internal": InternalRequestsClient()})

    with baseplate.server_context("test") as context:
        setattr(context, "raw_edge_context", b"test payload")

        response = context.internal.get(http_server.url)

        assert response.status_code == 204
        assert response.text == ""
        assert http_server.requests[0].method == "GET"
        assert http_server.requests[0].span.trace_id == context.span.trace_id
        assert http_server.requests[0].span.parent_id == context.span.id
        assert http_server.requests[0].span.id != context.span.id
        assert http_server.requests[0].raw_edge_context == b"test payload"


def test_external_client_doesnt_send_headers(http_server):
    baseplate = Baseplate(
        {"external.filter.ip_allowlist": "127.0.0.0/8", "external.filter.port_denylist": "0"}
    )
    baseplate.configure_context({"external": ExternalRequestsClient()})

    with baseplate.server_context("test") as context:
        setattr(context, "raw_edge_context", b"test payload")

        response = context.external.get(http_server.url)

        assert response.status_code == 204
        assert response.text == ""
        assert http_server.requests[0].method == "GET"
        assert "X-Trace" not in http_server.requests[0].headers
        assert "X-Parent" not in http_server.requests[0].headers
        assert "X-Span" not in http_server.requests[0].headers
        assert "X-Edge-Request" not in http_server.requests[0].headers


@pytest.mark.parametrize("client_cls", [InternalRequestsClient, ExternalRequestsClient])
@pytest.mark.parametrize("client_name", [None, "", "complex.client$name"])
@pytest.mark.parametrize("method", ["DELETE", "GET", "HEAD", "OPTIONS", "PATCH", "PUT", "POST"])
def test_prometheus_http_client_metrics(client_cls, client_name, method, http_server):
    baseplate = Baseplate(
        {"myclient.filter.ip_allowlist": "127.0.0.0/8", "myclient.filter.port_denylist": "0"}
    )
    logger.info("Created baseplate.")
    if client_name is None:
        baseplate.configure_context({"myclient": client_cls()})
    else:
        baseplate.configure_context({"myclient": client_cls(client_name=client_name)})
    logger.info(
        f"Configured baseplate context with requests client. [type={client_cls.__name__}, client_name={client_name}]"
    )

    observer = PrometheusBaseplateObserver()
    logger.info(f"Created observer. [type={observer.__class__.__name__}]")
    baseplate.register(observer)
    logger.info("Registered observer with baseplate.")

    # we need to clear metrics between test runs from the parametrize annotations
    REGISTRY._names_to_collectors["http_client_requests_total"].clear()
    REGISTRY._names_to_collectors["http_client_latency_seconds"].clear()
    REGISTRY._names_to_collectors["http_client_active_requests"].clear()

    with baseplate.server_context("test") as context:
        logger.debug("Created server context.")
        fn = getattr(context.myclient, method.lower())
        logger.debug(f"Making HTTP call. [method={method}, endpoint={http_server.url}]")
        response = fn(http_server.url)
        logger.debug(f"Made HTTP call. [method={method}, endpoint={http_server.url}]")

    assert response.status_code == 204

    http_client_requests_total = REGISTRY._names_to_collectors[
        "http_client_requests_total"
    ].collect()
    assert len(http_client_requests_total) == 1
    assert len(http_client_requests_total[0].samples) == 2  # _total and _created
    http_client_requests_total_total = http_client_requests_total[0].samples[0]
    assert http_client_requests_total_total.name == "http_client_requests_total"
    assert http_client_requests_total_total.labels.get("http_method") == method
    assert http_client_requests_total_total.labels.get("http_response_code") == "204"
    assert http_client_requests_total_total.labels.get("http_success") == "true"
    assert http_client_requests_total_total.labels.get("http_slug") == (
        client_name if client_name is not None else "myclient"
    )
    assert http_client_requests_total_total.value == 1

    http_client_latency_seconds = REGISTRY._names_to_collectors[
        "http_client_latency_seconds"
    ].collect()
    assert len(http_client_latency_seconds) == 1
    assert (
        len(http_client_latency_seconds[0].samples) == 18
    )  # 15 time buckets (inc. inf), _count, _sum and _created
    http_client_latency_seconds_inf = http_client_latency_seconds[0].samples[14]
    assert http_client_latency_seconds_inf.labels.get("le") == "+Inf"
    assert http_client_latency_seconds_inf.labels.get("http_method") == method
    assert http_client_latency_seconds_inf.labels.get("http_success") == "true"
    assert http_client_latency_seconds_inf.labels.get("http_slug") == (
        client_name if client_name is not None else "myclient"
    )
    assert http_client_latency_seconds_inf.value == 1
    http_client_latency_seconds_count = http_client_latency_seconds[0].samples[
        15
    ]  # no need to test the tags again since they all in the same sample
    assert http_client_latency_seconds_count.value == 1

    http_client_active_requests = REGISTRY._names_to_collectors[
        "http_client_active_requests"
    ].collect()
    assert len(http_client_active_requests) == 1
    assert len(http_client_active_requests[0].samples) == 1
    http_client_active_requests_samples = http_client_active_requests[0].samples[0]
    assert http_client_active_requests_samples.labels.get("http_method") == method
    assert http_client_active_requests_samples.labels.get("http_slug") == (
        client_name if client_name is not None else "myclient"
    )
    assert http_client_active_requests_samples.value == 0


@pytest.mark.parametrize("method", ["DELETE", "GET", "HEAD", "OPTIONS", "PATCH", "PUT", "POST"])
def test_prometheus_http_client_metrics_inside_local_span(method, http_server):
    baseplate = Baseplate(
        {"myclient.filter.ip_allowlist": "127.0.0.0/8", "myclient.filter.port_denylist": "0"}
    )
    logger.info("Created baseplate.")
    baseplate.configure_context({"myclient": InternalRequestsClient()})

    observer = PrometheusBaseplateObserver()
    logger.info(f"Created observer. [type={observer.__class__.__name__}]")
    baseplate.register(observer)
    logger.info("Registered observer with baseplate.")

    # we need to clear metrics between test runs from the parametrize annotations
    REGISTRY._names_to_collectors["http_client_requests_total"].clear()
    REGISTRY._names_to_collectors["http_client_latency_seconds"].clear()
    REGISTRY._names_to_collectors["http_client_active_requests"].clear()

    with baseplate.server_context("test") as context:
        with context.span.make_child("local", local=True) as span:
            logger.debug("Created local context.")
            fn = getattr(span.context.myclient, method.lower())
            logger.debug(f"Making HTTP call. [method={method}, endpoint={http_server.url}]")
            response = fn(http_server.url)
            logger.debug(f"Made HTTP call. [method={method}, endpoint={http_server.url}]")

    assert response.status_code == 204

    # Only making quick checks here, since this is meant to ensure that we do record metrics inside
    # a local span at all. Whether tags etc. are valid are tested in a previous test:
    # test_prometheus_http_client_metrics
    http_client_requests_total = REGISTRY._names_to_collectors[
        "http_client_requests_total"
    ].collect()
    assert len(http_client_requests_total) == 1
    assert len(http_client_requests_total[0].samples) == 2  # _total and _created
    http_client_requests_total_total = http_client_requests_total[0].samples[0]
    assert http_client_requests_total_total.value == 1

    http_client_latency_seconds = REGISTRY._names_to_collectors[
        "http_client_latency_seconds"
    ].collect()
    assert len(http_client_latency_seconds) == 1
    assert (
        len(http_client_latency_seconds[0].samples) == 18
    )  # 15 time buckets (inc. inf), _count, _sum and _created
    http_client_latency_seconds_inf = http_client_latency_seconds[0].samples[14]
    assert http_client_latency_seconds_inf.labels.get("le") == "+Inf"
    assert http_client_latency_seconds_inf.value == 1
    http_client_latency_seconds_count = http_client_latency_seconds[0].samples[15]
    assert http_client_latency_seconds_count.value == 1

    http_client_active_requests = REGISTRY._names_to_collectors[
        "http_client_active_requests"
    ].collect()
    assert len(http_client_active_requests) == 1
    assert len(http_client_active_requests[0].samples) == 1
    http_client_active_requests_samples = http_client_active_requests[0].samples[0]
    assert http_client_active_requests_samples.value == 0



import logging
import time

from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from baseplate import _ExcInfo
from baseplate import BaseplateObserver
from baseplate import LocalSpan
from baseplate import RequestContext
from baseplate import Span
from baseplate import SpanObserver
from baseplate.lib.prometheus_metrics import PrometheusHTTPClientMetrics
from baseplate.lib.prometheus_metrics import PrometheusHTTPServerMetrics
from baseplate.lib.prometheus_metrics import PrometheusThriftClientMetrics
from baseplate.lib.prometheus_metrics import PrometheusThriftServerMetrics
from baseplate.thrift.ttypes import Error
from baseplate.thrift.ttypes import ErrorCode


NANOSECONDS_PER_SECOND = 1e9

logger = logging.getLogger(__name__)


class PrometheusBaseplateObserver(BaseplateObserver):
    """Metrics collecting observer.

    This observer reports Prometheus metrics for HTTP or Thrift client and servers.

    There must be a "protocol" tag set on the server_span that indictes the protocol, either "http" or "thrift",
    i.e. `span.set_tag("protocol", "http")` or `span.set_tag("protocol", "thrift")`.

    """

    def __init__(self) -> None:
        pass

    def on_server_span_created(self, context: RequestContext, server_span: Span) -> None:
        observer = PrometheusServerSpanObserver()
        server_span.register(observer)


class PrometheusServerSpanObserver(SpanObserver):
    def __init__(self) -> None:
        self.tags: Dict[str, Any] = {}
        self.start_time: Optional[int] = None
        self.metrics: Optional[
            Union[
                PrometheusHTTPServerMetrics,
                PrometheusThriftServerMetrics,
            ]
        ] = None

    def on_set_tag(self, key: str, value: Any) -> None:
        self.tags[key] = value

    def get_prefix(self) -> str:
        return f"{self.protocol}_server"

    def get_labels(self) -> Dict[str, str]:
        return self.tags

    @property
    def protocol(self) -> str:
        return self.tags.get("protocol", "unknown")

    def set_metrics_by_protocol(self) -> None:
        if self.protocol == "thrift":
            self.metrics = PrometheusThriftServerMetrics()
        elif self.protocol == "http":
            self.metrics = PrometheusHTTPServerMetrics()
        else:
            logger.warning(
                "No valid protocol set for Prometheus metric collection, metrics won't be collected. Expected 'http' or 'thrift' protocol. Actual protocol: %s",
                self.protocol,
            )

    def on_start(self) -> None:
        self.set_metrics_by_protocol()
        if self.metrics is None:
            logger.warning(
                "No metrics set for Prometheus metric collection. Metrics will not be exported correctly."
            )
            return
        self.start_time = time.perf_counter_ns()
        self.metrics.active_requests_metric(self.tags).inc()

    def on_incr_tag(self, key: str, delta: float) -> None:
        pass

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        if self.metrics is None:
            logger.warning(
                "No metrics set for Prometheus metric collection. Metrics will not be exported correctly."
            )
            return

        if exc_info is not None:
            self.tags["exception_type"] = exc_info[1].__class__.__name__
            self.tags["success"] = "false"

        if self.tags.get("exception_type", "") == "":
            self.tags["success"] = "true"

        self.metrics.active_requests_metric(self.tags).dec()
        self.metrics.requests_total_metric(self.tags).inc()

        if self.start_time is not None:
            elapsed_ns = time.perf_counter_ns() - self.start_time
            self.metrics.latency_seconds_metric(self.tags).observe(
                elapsed_ns / NANOSECONDS_PER_SECOND
            )

        if isinstance(self.metrics, PrometheusHTTPServerMetrics):
            self.metrics.response_size_bytes_metric(self.tags).observe(
                self.tags.get("http.response_length") or 0
            )

            self.metrics.request_size_bytes_metric(self.tags).observe(
                self.tags.get("http.request_length") or 0
            )

    def on_child_span_created(self, span: Span) -> None:
        observer: Any = None
        if isinstance(span, LocalSpan):
            observer = PrometheusLocalSpanObserver()
        else:
            observer = PrometheusClientSpanObserver()

        span.register(observer)


class PrometheusClientSpanObserver(SpanObserver):
    def __init__(self) -> None:
        self.tags: Dict[str, Any] = {}
        self.start_time: Optional[int] = None
        self.metrics: Optional[
            Union[PrometheusHTTPClientMetrics, PrometheusThriftClientMetrics]
        ] = None

    def on_set_tag(self, key: str, value: Any) -> None:
        self.tags[key] = value

    def get_prefix(self) -> str:
        return f"{self.protocol}_client"

    def get_labels(self) -> Dict[str, str]:
        return self.tags

    @property
    def protocol(self) -> str:
        return self.tags.get("protocol", "unknown")

    def set_metrics_by_protocol(self) -> None:
        if self.protocol == "http":
            self.metrics = PrometheusHTTPClientMetrics()
        elif self.protocol == "thrift":
            self.metrics = PrometheusThriftClientMetrics()
        else:
            logger.warning(
                "No valid protocol set for Prometheus metric collection, metrics won't be collected. Expected 'http' or 'thrift' protocol. Actual protocol: %s",
                self.protocol,
            )
        return

    def on_start(self) -> None:
        self.set_metrics_by_protocol()
        if self.metrics is None:
            logger.warning(
                "No metrics set for Prometheus metric collection. Metrics will not be exported correctly."
            )
            return
        self.start_time = time.perf_counter_ns()
        self.metrics.active_requests_metric(self.tags).inc()

    def on_incr_tag(self, key: str, delta: float) -> None:
        pass

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        if self.metrics is None:
            logger.warning(
                "No metrics set for Prometheus metric collection. Metrics will not be exported correctly."
            )
            return

        self.tags["success"] = "true"
        if exc_info is not None:
            exc = exc_info[1]
            self.tags["exception_type"] = exc.__class__.__name__
            self.tags["success"] = "false"
            if self.protocol == "thrift" and isinstance(exc, Error):
                self.tags["thrift_status_code"] = exc.code
                self.tags["thrift_status"] = ErrorCode()._VALUES_TO_NAMES.get(exc.code, "")

        self.metrics.active_requests_metric(self.tags).dec()
        self.metrics.requests_total_metric(self.tags).inc()
        if self.start_time is not None:
            elapsed_ns = time.perf_counter_ns() - self.start_time
            self.metrics.latency_seconds_metric(self.tags).observe(
                elapsed_ns / NANOSECONDS_PER_SECOND
            )

    def on_child_span_created(self, span: Span) -> None:
        observer: Optional[SpanObserver] = None
        if isinstance(span, LocalSpan):
            observer = PrometheusLocalSpanObserver()
        else:
            observer = PrometheusClientSpanObserver()

        span.register(observer)


class PrometheusLocalSpanObserver(SpanObserver):
    def __init__(self) -> None:
        logger.debug("PrometheusLocalSpanObserver not implemented")

    # Proper implementation for PrometheusLocalSpanObserver will come in a future PR
    # In the meantime, we need this logic in place to be able to emit http client metrics
    # when running inside a local span
    def on_child_span_created(self, span: Span) -> None:
        observer: Optional[SpanObserver] = None
        if isinstance(span, LocalSpan):
            observer = PrometheusLocalSpanObserver()
        else:
            observer = PrometheusClientSpanObserver()

        span.register(observer)



import unittest

import fakeredis

from baseplate.clients.redis_cluster import HotKeyTracker


class HotKeyTrackerTests(unittest.TestCase):
    def setUp(self):
        self.rc = fakeredis.FakeStrictRedis()

    def test_increment_reads_once(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        tracker.increment_keys_read_counter(["foo"], ignore_errors=False)
        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [(b"foo", float(1))],
        )

    def test_increment_several_reads(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        for _ in range(5):
            tracker.increment_keys_read_counter(["foo"], ignore_errors=False)

        tracker.increment_keys_read_counter(["bar"], ignore_errors=False)

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [(b"bar", float(1)), (b"foo", float(5))],
        )

    def test_reads_disabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 0, 0)
        for _ in range(5):
            tracker.maybe_track_key_usage(["GET", "foo"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [],
        )

    def test_reads_enabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        for _ in range(5):
            tracker.maybe_track_key_usage(["GET", "foo"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-reads", "-inf", "+inf", withscores=True
            ),
            [(b"foo", float(5))],
        )

    def test_writes_enabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 1, 1)
        for _ in range(5):
            tracker.maybe_track_key_usage(["SET", "foo", "bar"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [(b"foo", float(5))],
        )

    def test_writes_disabled_tracking(self):
        tracker = HotKeyTracker(self.rc, 0, 0)
        for _ in range(5):
            tracker.maybe_track_key_usage(["SET", "foo", "bar"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [],
        )

    def test_write_multikey_commands(self):
        tracker = HotKeyTracker(self.rc, 1, 1)

        tracker.maybe_track_key_usage(["DEL", "foo", "bar"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [(b"bar", float(1)), (b"foo", float(1))],
        )

    def test_write_batchkey_commands(self):
        tracker = HotKeyTracker(self.rc, 1, 1)

        tracker.maybe_track_key_usage(["MSET", "foo", "bar", "baz", "wednesday"])

        self.assertEqual(
            tracker.redis_client.zrangebyscore(
                "baseplate-hot-key-tracker-writes", "-inf", "+inf", withscores=True
            ),
            [(b"baz", float(1)), (b"foo", float(1))],
        )



from math import ceil
from typing import Any
from typing import Dict
from typing import Optional

import redis

from prometheus_client import Gauge

# redis.client.StrictPipeline was renamed to redis.client.Pipeline in version 3.0
try:
    from redis.client import StrictPipeline as Pipeline  # type: ignore
except ImportError:
    from redis.client import Pipeline

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib import message_queue
from baseplate.lib import metrics


def pool_from_config(
    app_config: config.RawConfig, prefix: str = "redis.", **kwargs: Any
) -> redis.ConnectionPool:
    """Make a ConnectionPool from a configuration dictionary.

    The keys useful to :py:func:`pool_from_config` should be prefixed, e.g.
    ``redis.url``, ``redis.max_connections``, etc. The ``prefix`` argument
    specifies the prefix used to filter keys.  Each key is mapped to a
    corresponding keyword argument on the :py:class:`redis.ConnectionPool`
    constructor.

    Supported keys:

    * ``url`` (required): a URL like ``redis://localhost/0``.
    * ``max_connections``: an integer maximum number of connections in the pool
    * ``socket_connect_timeout``: how long to wait for sockets to connect. e.g.
        ``200 milliseconds`` (:py:func:`~baseplate.lib.config.Timespan`)
    * ``socket_timeout``: how long to wait for socket operations, e.g.
        ``200 milliseconds`` (:py:func:`~baseplate.lib.config.Timespan`)

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "url": config.String,
            "max_connections": config.Optional(config.Integer, default=None),
            "socket_connect_timeout": config.Optional(config.Timespan, default=None),
            "socket_timeout": config.Optional(config.Timespan, default=None),
        }
    )
    options = parser.parse(prefix[:-1], app_config)

    if options.max_connections is not None:
        kwargs.setdefault("max_connections", options.max_connections)
    if options.socket_connect_timeout is not None:
        kwargs.setdefault("socket_connect_timeout", options.socket_connect_timeout.total_seconds())
    if options.socket_timeout is not None:
        kwargs.setdefault("socket_timeout", options.socket_timeout.total_seconds())

    return redis.BlockingConnectionPool.from_url(options.url, **kwargs)


class RedisClient(config.Parser):
    """Configure a Redis client.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`pool_from_config` for available configuration settings.

    """

    def __init__(self, **kwargs: Any):
        self.kwargs = kwargs

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "RedisContextFactory":
        connection_pool = pool_from_config(raw_config, f"{key_path}.", **self.kwargs)
        return RedisContextFactory(connection_pool, key_path)


class RedisContextFactory(ContextFactory):
    """Redis client context factory.

    This factory will attach a
    :py:class:`~baseplate.clients.redis.MonitoredRedisConnection` to an
    attribute on the :py:class:`~baseplate.RequestContext`. When Redis commands
    are executed via this connection object, they will use connections from the
    provided :py:class:`redis.ConnectionPool` and automatically record
    diagnostic information.

    :param connection_pool: A connection pool.

    """

    PROM_PREFIX = "bp_redis_pool"
    PROM_LABELS = ["pool"]

    max_connections = Gauge(
        f"{PROM_PREFIX}_max_size",
        "Maximum number of connections allowed in this redisbp pool",
        PROM_LABELS,
    )
    idle_connections = Gauge(
        f"{PROM_PREFIX}_idle_connections",
        "Number of idle connections in this redisbp pool",
        PROM_LABELS,
    )
    open_connections = Gauge(
        f"{PROM_PREFIX}_active_connections",
        "Number of open connections in this redisbp pool",
        PROM_LABELS,
    )

    def __init__(self, connection_pool: redis.ConnectionPool, name: str = "redis"):
        self.connection_pool = connection_pool

        if isinstance(connection_pool, redis.BlockingConnectionPool):
            self.max_connections.labels(name).set_function(lambda: connection_pool.max_connections)
            self.idle_connections.labels(name).set_function(connection_pool.pool.qsize)
            self.open_connections.labels(name).set_function(
                lambda: len(connection_pool._connections)  # type: ignore
            )

    def report_runtime_metrics(self, batch: metrics.Client) -> None:
        if not isinstance(self.connection_pool, redis.BlockingConnectionPool):
            return

        size = self.connection_pool.max_connections
        open_connections = len(self.connection_pool._connections)  # type: ignore
        available = self.connection_pool.pool.qsize()
        in_use = size - available

        batch.gauge("pool.size").replace(size)
        batch.gauge("pool.in_use").replace(in_use)
        batch.gauge("pool.open_and_available").replace(open_connections - in_use)

    def make_object_for_context(self, name: str, span: Span) -> "MonitoredRedisConnection":
        return MonitoredRedisConnection(name, span, self.connection_pool)


# pylint: disable=too-many-public-methods
class MonitoredRedisConnection(redis.StrictRedis):
    """Redis connection that collects diagnostic information.

    This connection acts like :py:class:`redis.StrictRedis` except that all
    operations are automatically wrapped with diagnostic collection.

    The interface is the same as that class except for the
    :py:meth:`~baseplate.clients.redis.MonitoredRedisConnection.pipeline`
    method.

    """

    def __init__(self, context_name: str, server_span: Span, connection_pool: redis.ConnectionPool):
        self.context_name = context_name
        self.server_span = server_span

        super().__init__(connection_pool=connection_pool)

    def execute_command(self, *args: Any, **kwargs: Any) -> Any:
        command = args[0]
        trace_name = f"{self.context_name}.{command}"

        with self.server_span.make_child(trace_name):
            return super().execute_command(command, *args[1:], **kwargs)

    # pylint: disable=arguments-differ
    def pipeline(  # type: ignore
        self, name: str, transaction: bool = True, shard_hint: Optional[str] = None
    ) -> "MonitoredRedisPipeline":
        """Create a pipeline.

        This returns an object on which you can call the standard Redis
        commands. Execution will be deferred until ``execute`` is called. This
        is useful for saving round trips.

        :param name: The name to attach to diagnostics for this pipeline.
        :param transaction: Whether or not the commands in the pipeline
            are wrapped with a transaction and executed atomically.

        """
        return MonitoredRedisPipeline(
            f"{self.context_name}.pipeline_{name}",
            self.server_span,
            self.connection_pool,
            self.response_callbacks,
            transaction=transaction,
            shard_hint=shard_hint,
        )

    # these commands are not yet implemented, but probably not unimplementable
    def transaction(self, *args: Any, **kwargs: Any) -> Any:
        """Not currently implemented."""
        raise NotImplementedError


class MonitoredRedisPipeline(Pipeline):
    def __init__(
        self,
        trace_name: str,
        server_span: Span,
        connection_pool: redis.ConnectionPool,
        response_callbacks: Dict,
        **kwargs: Any,
    ):
        self.trace_name = trace_name
        self.server_span = server_span
        super().__init__(connection_pool, response_callbacks, **kwargs)

    # pylint: disable=arguments-differ
    def execute(self, **kwargs: Any) -> Any:
        with self.server_span.make_child(self.trace_name):
            return super().execute(**kwargs)


class MessageQueue:
    """A Redis-backed variant of :py:class:`~baseplate.lib.message_queue.MessageQueue`.

    :param name: can be any string.

    :param client: should be a :py:class:`redis.ConnectionPool` or
           :py:class:`redis.BlockingConnectionPool` from which a client
           connection can be created from (preferably generated from the
           :py:func:`pool_from_config` helper).

    """

    def __init__(self, name: str, client: redis.ConnectionPool):
        self.queue = name
        if isinstance(client, (redis.BlockingConnectionPool, redis.ConnectionPool)):
            self.client = redis.Redis(connection_pool=client)
        else:
            self.client = client

    def get(self, timeout: Optional[float] = None) -> bytes:
        """Read a message from the queue.

        :param timeout: If the queue is empty, the call will block up to
            ``timeout`` seconds or forever if ``None``, if a float is given,
            it will be rounded up to be an integer
        :raises: :py:exc:`~baseplate.lib.message_queue.TimedOutError` The queue
            was empty for the allowed duration of the call.

        """
        if isinstance(timeout, float):
            timeout = int(ceil(timeout))

        if timeout == 0:
            message = self.client.lpop(self.queue)
        else:
            message = self.client.blpop(self.queue, timeout=timeout or 0)

            if message:
                message = message[1]

        if not message:
            raise message_queue.TimedOutError

        return message

    def put(  # pylint: disable=unused-argument
        self, message: bytes, timeout: Optional[float] = None
    ) -> None:
        """Add a message to the queue.

        :param message: will be typecast to a string upon storage and will come
               out of the queue as a string regardless of what type they are
               when passed into this method.
        """
        self.client.rpush(self.queue, message)

    def unlink(self) -> None:
        """Not implemented for Redis variant."""

    def close(self) -> None:
        """Close queue when finished.

        Will delete the queue from the Redis server (Note, can still enqueue
        and dequeue as the actions will recreate the queue)
        """
        self.client.delete(self.queue)



import unittest

from baseplate.clients import thrift
from baseplate.thrift import BaseplateServiceV2


class EnumerateServiceMethodsTests(unittest.TestCase):
    def test_enumerate_none(self):
        class Iface:
            pass

        class ExampleClient(Iface):
            pass

        methods = list(thrift._enumerate_service_methods(ExampleClient))

        self.assertEqual(methods, [])

    def test_enumerate_some(self):
        class Iface:
            def some_method(self):
                pass

        class ExampleClient(Iface):
            pass

        methods = list(thrift._enumerate_service_methods(ExampleClient))

        self.assertEqual(set(methods), {"some_method"})

    def test_inherited(self):
        class Iface:
            def local_method(self):
                pass

        class ExampleClient(BaseplateServiceV2.Client, Iface):
            pass

        methods = list(thrift._enumerate_service_methods(ExampleClient))

        self.assertEqual(set(methods), {"is_healthy", "local_method"})

    def test_not_subclass_of_iface(self):
        class ExampleClient:
            pass

        with self.assertRaises(AssertionError):
            list(thrift._enumerate_service_methods(ExampleClient))



"""Watch nodes in ZooKeeper and sync their contents to disk on change."""
import argparse
import configparser
import json
import logging
import os
import sys
import time

from pathlib import Path
from typing import Any
from typing import NoReturn
from typing import Optional

from kazoo.client import KazooClient
from kazoo.protocol.states import ZnodeStat

from baseplate.lib import config
from baseplate.lib.live_data.zookeeper import zookeeper_client_from_config
from baseplate.lib.secrets import secrets_store_from_config
from baseplate.server import EnvironmentInterpolation

try:
    import boto3  # type: ignore
    from botocore.client import ClientError  # type: ignore

    S3_FETCHER_ENABLED = True
except ImportError:
    S3_FETCHER_ENABLED = False


logger = logging.getLogger(__name__)


HEARTBEAT_INTERVAL = 300


class NodeWatcher:
    def __init__(self, dest: str, owner: int, group: int, mode: int):
        self.dest = dest
        self.owner = owner
        self.group = group
        self.mode = mode

    @staticmethod
    def get_encrypted_data_from_s3(
        bucket_name: str, file_key: str, region_name: str, sse_key: str
    ) -> Optional[bytes]:
        s3_client = boto3.client(
            "s3",
            region_name=region_name,
        )
        data = None
        # The S3 data must be Server Side Encrypted and we should have the decryption key.
        kwargs = {
            "Bucket": bucket_name,
            "Key": file_key,
            "SSECustomerKey": sse_key,
            "SSECustomerAlgorithm": "AES256",
        }
        try:
            s3_object = s3_client.get_object(**kwargs)
            # Returns bytes.
            data = s3_object["Body"].read()
        except ClientError as error:
            logger.exception(
                "Failed to update live config: failed to load data from S3. Received error code %s: %s",
                error.response["Error"]["Code"],
                error.response["Error"]["Message"],
            )
            return None
        except ValueError as error:
            logger.exception(
                "Failed to update live config: params for loading from S3 are incorrect. Received error: %s",
                error,
            )
            return None
        return data

    @staticmethod
    def get_data_to_write(json_data: dict, data: Optional[bytes]) -> Optional[bytes]:
        # Check if we have a JSON in a special format containing:
        # data = {
        #    "live_data_watcher_load_type": str
        # }
        # If the load type is 'S3', this format is an indication that we support
        # downloading the contents of encrypted files uploaded to S3.
        if json_data.get("live_data_watcher_load_type") == "S3":
            if not S3_FETCHER_ENABLED:
                logger.error("Attempted to use s3 fetcher but packages are missing")
                return None

            try:
                bucket_name = json_data["bucket_name"]
                file_key = json_data["file_key"]
                sse_key = json_data["sse_key"]
                region_name = json_data["region_name"]
            except KeyError:
                # We require all of these keys to properly read from S3.
                logger.exception(
                    "Failed to update live config: unble to fetch content from s3: source config has invalid or missing keys."
                )
                return None
            # If we have all the correct keys, attempt to read the config from S3.
            data = NodeWatcher.get_encrypted_data_from_s3(
                bucket_name=bucket_name, file_key=file_key, region_name=region_name, sse_key=sse_key
            )
        return data

    def handle_empty_data(self) -> None:
        # the data node does not exist
        try:
            logger.info("Removing %r; watched node deleted.", self.dest)
            os.unlink(self.dest)
        except OSError as exc:
            logger.debug("%s: couldn't unlink: %s", self.dest, exc)

    def on_change(self, data: Optional[bytes], _znode_stat: ZnodeStat) -> None:
        if data is None:
            self.handle_empty_data()
            return

        # swap out the file atomically so clients watching the file never catch
        # us mid-write.
        logger.info("Updating %r", self.dest)
        Path(self.dest).parent.mkdir(parents=True, exist_ok=True)
        with open(self.dest + ".tmp", "wb") as tmpfile:
            if self.owner and self.group:
                os.fchown(tmpfile.fileno(), self.owner, self.group)
            os.fchmod(tmpfile.fileno(), self.mode)
            try:
                json_data = json.loads(data.decode("UTF-8"))
            except json.decoder.JSONDecodeError:
                # If JSON fails to decode, still write the bytes data since
                # we don't necessarily know if the the contents of the znode
                # had to be valid JSON anyways.
                tmpfile.write(data)
            else:
                # If no exceptions, we have valid JSON, and can parse accordingly.
                data = NodeWatcher.get_data_to_write(json_data, data)
                if data is None:
                    logger.warning(
                        "No data written to destination file. Something is likely misconfigured."
                    )
                    return
                tmpfile.write(data)
        os.rename(self.dest + ".tmp", self.dest)


def watch_zookeeper_nodes(zookeeper: KazooClient, nodes: Any) -> NoReturn:
    for node in nodes:
        watcher = NodeWatcher(node.dest, node.owner, node.group, node.mode)
        zookeeper.DataWatch(node.source, watcher.on_change)

    # all the interesting stuff is now happening in the Kazoo worker thread
    # and so we'll just spin and periodically heartbeat to prove we're alive.
    while True:
        time.sleep(HEARTBEAT_INTERVAL)

        # see the comment in baseplate.live_data.zookeeper for explanation of
        # how reconnects work with the background thread.
        if zookeeper.connected:
            for node in nodes:
                try:
                    logger.debug("Heartbeating %s", node.dest)

                    # this will make FileWatchers re-parse the file on the next
                    # read which is unfortunate but we do it anyway. it's
                    # important to monitor that the file is being updated as
                    # accurately as possible rather than using a separate file
                    # or mechanism as a proxy. for example, the actual output
                    # file could have bogus permissions that would go unnoticed
                    # if the heartbeat still worked independently.
                    os.utime(node.dest, None)
                except OSError as exc:
                    logger.warning("%s: could not heartbeat: %s", node.dest, exc)


def main() -> NoReturn:
    arg_parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__)
    arg_parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )
    arg_parser.add_argument(
        "--debug", default=False, action="store_true", help="enable debug logging"
    )
    args = arg_parser.parse_args()

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level, format="%(message)s")

    # quiet kazoo's verbose logs a bit
    logging.getLogger("kazoo").setLevel(logging.WARNING)

    parser = configparser.RawConfigParser(interpolation=EnvironmentInterpolation())
    parser.read_file(args.config_file)
    watcher_config = dict(parser.items("live-data"))

    cfg = config.parse_config(
        watcher_config,
        {
            "nodes": config.DictOf(
                {
                    "source": config.String,
                    "dest": config.String,
                    "owner": config.Optional(config.UnixUser),
                    "group": config.Optional(config.UnixGroup),
                    "mode": config.Optional(config.Integer(base=8), default=0o400),  # type: ignore
                }
            )
        },
    )
    # pylint: disable=maybe-no-member
    nodes = cfg.nodes.values()

    secrets = secrets_store_from_config(watcher_config, timeout=30)
    zookeeper = zookeeper_client_from_config(secrets, watcher_config, read_only=True)
    zookeeper.start()
    try:
        watch_zookeeper_nodes(zookeeper, nodes)
    finally:
        zookeeper.stop()


if __name__ == "__main__":
    main()



import socket

from queue import Queue
from unittest import mock

import kombu
import pytest

from gevent.server import StreamServer

from baseplate import Baseplate
from baseplate import RequestContext
from baseplate import ServerSpan
from baseplate.frameworks.queue_consumer.kombu import FatalMessageHandlerError
from baseplate.frameworks.queue_consumer.kombu import KombuConsumerWorker
from baseplate.frameworks.queue_consumer.kombu import KombuMessageHandler
from baseplate.frameworks.queue_consumer.kombu import KombuQueueConsumerFactory

from .... import does_not_raise


@pytest.fixture
def context():
    return mock.Mock(spec=RequestContext)


@pytest.fixture
def span():
    return mock.Mock(spec=ServerSpan)


@pytest.fixture
def baseplate(context, span):
    bp = mock.MagicMock(spec=Baseplate)
    bp._metrics_client = None
    bp.make_server_span().__enter__.return_value = span
    bp.make_context_object.return_value = context
    # Reset the mock calls since setting up the span actually triggers a "call"
    # to bp.make_server_span
    bp.reset_mock()
    return bp


@pytest.fixture
def name():
    return "test-handler"


class TestKombuMessageHandler:
    @pytest.fixture
    def message(self):
        msg = mock.Mock(spec=kombu.Message)
        msg.delivery_info = {
            "routing_key": "routing-key",
            "consumer_tag": "consumer-tag",
            "delivery_tag": "delivery-tag",
            "exchange": "exchange",
        }
        msg.decode.return_value = {"foo": "bar"}
        return msg

    def test_handle(self, context, span, baseplate, name, message):
        handler_fn = mock.Mock()
        handler = KombuMessageHandler(baseplate, name, handler_fn)
        handler.handle(message)
        baseplate.make_context_object.assert_called_once()
        baseplate.make_server_span.assert_called_once_with(context, name)
        baseplate.make_server_span().__enter__.assert_called_once()
        span.set_tag.assert_has_calls(
            [
                mock.call("kind", "consumer"),
                mock.call("amqp.routing_key", "routing-key"),
                mock.call("amqp.consumer_tag", "consumer-tag"),
                mock.call("amqp.delivery_tag", "delivery-tag"),
                mock.call("amqp.exchange", "exchange"),
            ],
            any_order=True,
        )
        handler_fn.assert_called_once_with(context, message.decode(), message)
        message.ack.assert_called_once()

    @pytest.mark.parametrize(
        "err,expectation",
        [
            (ValueError(), does_not_raise()),
            (FatalMessageHandlerError(), pytest.raises(FatalMessageHandlerError)),
        ],
    )
    def test_errors(self, err, expectation, baseplate, name, message):
        def handler_fn(ctx, body, msg):
            raise err

        handler = KombuMessageHandler(baseplate, name, handler_fn)
        with expectation:
            handler.handle(message)
        message.requeue.assert_called_once()

    @pytest.mark.parametrize(
        "err,expectation",
        [
            (ValueError(), does_not_raise()),
            (FatalMessageHandlerError(), pytest.raises(FatalMessageHandlerError)),
        ],
    )
    def test_errors_with_error_handler_fn(
        self, err, expectation, context, baseplate, name, message
    ):
        def handler_fn(ctx, body, msg):
            raise err

        error_handler_fn = mock.Mock()

        handler = KombuMessageHandler(baseplate, name, handler_fn, error_handler_fn)
        with expectation:
            handler.handle(message)
        error_handler_fn.assert_called_once_with(context, message.decode(), message, err)
        message.ack.assert_not_called()
        message.requeue.assert_not_called()


@pytest.fixture
def connection():
    return mock.Mock(spec=kombu.Connection)


@pytest.fixture
def exchange():
    return mock.Mock(spec=kombu.Exchange)


@pytest.fixture
def routing_keys():
    return ["rk-1", "rk-2"]


class TestQueueConsumerFactory:
    @pytest.fixture
    def make_queue_consumer_factory(self, baseplate, exchange, connection, name, routing_keys):
        def _make_queue_consumer_factory(health_check_fn=None):
            worker_kwargs = {"prefetch_limit": 1}
            return KombuQueueConsumerFactory.new(
                baseplate=baseplate,
                exchange=exchange,
                connection=connection,
                queue_name=name,
                routing_keys=routing_keys,
                handler_fn=lambda ctx, body, msg: True,
                error_handler_fn=lambda ctx, body, msg: True,
                health_check_fn=health_check_fn,
                worker_kwargs=worker_kwargs,
            )

        return _make_queue_consumer_factory

    def test_new(self, baseplate, exchange, connection, name, routing_keys):
        handler_fn = mock.Mock()
        error_handler_fn = mock.Mock()
        health_check_fn = mock.Mock()
        worker_kwargs = {"prefetch_limit": 1}
        factory = KombuQueueConsumerFactory.new(
            baseplate=baseplate,
            exchange=exchange,
            connection=connection,
            queue_name=name,
            routing_keys=routing_keys,
            handler_fn=handler_fn,
            error_handler_fn=error_handler_fn,
            health_check_fn=health_check_fn,
            worker_kwargs=worker_kwargs,
        )
        assert factory.baseplate == baseplate
        assert factory.connection == connection
        assert factory.name == name
        assert factory.handler_fn == handler_fn
        assert factory.error_handler_fn == error_handler_fn
        assert factory.health_check_fn == health_check_fn
        for routing_key, queue in zip(routing_keys, factory.queues):
            assert queue.routing_key == routing_key
            assert queue.name == name
            assert queue.exchange == exchange
        assert worker_kwargs == worker_kwargs

    def test_build_pump_worker(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        work_queue = Queue(maxsize=10)
        pump = factory.build_pump_worker(work_queue)
        assert isinstance(pump, KombuConsumerWorker)
        assert pump.connection == factory.connection
        assert pump.queues == factory.queues
        assert pump.work_queue == work_queue
        assert pump.kwargs == {"prefetch_limit": 1}

    def test_build_message_handler(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        handler = factory.build_message_handler()
        assert isinstance(handler, KombuMessageHandler)
        assert handler.baseplate == factory.baseplate
        assert handler.name == factory.name
        assert handler.handler_fn == factory.handler_fn
        assert handler.error_handler_fn == factory.error_handler_fn

    @pytest.mark.parametrize("health_check_fn", [None, lambda req: True])
    def test_build_health_checker(self, health_check_fn, make_queue_consumer_factory):
        factory = make_queue_consumer_factory(health_check_fn=health_check_fn)
        listener = mock.Mock(spec=socket.socket)
        health_checker = factory.build_health_checker(listener)
        assert isinstance(health_checker, StreamServer)


@pytest.fixture
def queues(name, exchange, routing_keys):
    return [kombu.Queue(name=name, exchange=exchange, routing_key=key) for key in routing_keys]


class TestKombuConsumerWorker:
    @pytest.fixture
    def consumer_worker(self, connection, queues):
        return KombuConsumerWorker(connection, queues, Queue(maxsize=10))

    def test_initial_state(self, consumer_worker):
        assert getattr(consumer_worker, "should_stop", None) is not True

    @mock.patch("baseplate.frameworks.queue_consumer.kombu.ConsumerMixin.run")
    def test_run(self, run, consumer_worker):
        consumer_worker.run()
        run.assert_called_once()

    def test_stop(self, consumer_worker):
        consumer_worker.stop()
        assert consumer_worker.should_stop is True



import unittest

try:
    import redis
except ImportError:
    raise unittest.SkipTest("redis-py is not installed")
else:
    del redis

from baseplate.lib.config import ConfigurationError
from baseplate.clients.redis import pool_from_config, RedisContextFactory


class PoolFromConfigTests(unittest.TestCase):
    def test_empty_config(self):
        with self.assertRaises(ConfigurationError):
            pool_from_config({})

    def test_basic_url(self):
        pool = pool_from_config({"redis.url": "redis://localhost:1234/0"})

        self.assertEqual(pool.connection_kwargs["host"], "localhost")
        self.assertEqual(pool.connection_kwargs["port"], 1234)
        self.assertEqual(pool.connection_kwargs["db"], 0)

    def test_metrics(self):
        max_connections = "123"
        ctx = RedisContextFactory(
            pool_from_config(
                {"redis.url": "redis://localhost:1234/0", "redis.max_connections": max_connections}
            )
        )
        metric = ctx.max_connections.collect()
        self.assertEqual(metric[0].samples[0].value, float(max_connections))

    def test_timeouts(self):
        pool = pool_from_config(
            {
                "redis.url": "redis://localhost:1234/0",
                "redis.socket_timeout": "30 seconds",
                "redis.socket_connect_timeout": "300 milliseconds",
            }
        )

        self.assertEqual(pool.connection_kwargs["socket_timeout"], 30)
        self.assertEqual(pool.connection_kwargs["socket_connect_timeout"], 0.3)

    def test_max_connections(self):
        pool = pool_from_config(
            {"redis.url": "redis://localhost:1234/0", "redis.max_connections": "300"}
        )

        self.assertEqual(pool.max_connections, 300)

    def test_kwargs_passthrough(self):
        pool = pool_from_config({"redis.url": "redis://localhost:1234/0"}, example="present")

        self.assertEqual(pool.connection_kwargs["example"], "present")

    def test_alternate_prefix(self):
        pool_from_config({"noodle.url": "redis://localhost:1234/0"}, prefix="noodle.")



import pytest

from baseplate.clients.kombu import connection_from_config
from baseplate.clients.kombu import KombuThriftSerializer
from baseplate.lib.config import ConfigurationError
from baseplate.testing.lib.secrets import FakeSecretsStore

from ... import does_not_raise
from ...integration.test_thrift.ttypes import ExampleStruct


def secrets():
    return FakeSecretsStore(
        {
            "secrets": {
                "secret/rabbitmq/account": {
                    "type": "credential",
                    "username": "spez",
                    "password": "hunter2",
                }
            },
        }
    )


@pytest.mark.parametrize(
    "app_config,kwargs,expectation,expected",
    [
        ({}, {}, pytest.raises(ConfigurationError), {}),
        ({"rabbitmq.virtual_host": "test"}, {}, pytest.raises(ConfigurationError), {}),
        ({"amqp.hostname": "amqp://rabbit.local:5672"}, {}, pytest.raises(ConfigurationError), {}),
        (
            {
                "rabbitmq.hostname": "amqp://rabbit.local:5672",
                "rabbitmq.credentials_secret": "secret/rabbitmq/account",
            },
            {},
            pytest.raises(ValueError),
            {},
        ),
        (
            {"rabbitmq.hostname": "amqp://rabbit.local:5672"},
            {},
            does_not_raise(),
            {"hostname": "rabbit.local", "userid": None, "password": None, "virtual_host": "/"},
        ),
        (
            {"rabbitmq.hostname": "amqp://rabbit.local:5672", "rabbitmq.virtual_host": "test"},
            {},
            does_not_raise(),
            {"hostname": "rabbit.local", "userid": None, "password": None, "virtual_host": "test"},
        ),
        (
            {"rabbitmq.hostname": "amqp://rabbit.local:5672", "rabbitmq.virtual_host": "test"},
            {"userid": "spez", "password": "hunter2"},
            does_not_raise(),
            {
                "hostname": "rabbit.local",
                "userid": "spez",
                "password": "hunter2",
                "virtual_host": "test",
            },
        ),
        (
            {
                "rabbitmq.hostname": "amqp://rabbit.local:5672",
                "rabbitmq.credentials_secret": "secret/rabbitmq/account",
            },
            {"secrets": secrets()},
            does_not_raise(),
            {
                "hostname": "rabbit.local",
                "userid": "spez",
                "password": "hunter2",
                "virtual_host": "/",
            },
        ),
    ],
)
def test_connection_from_config(app_config, kwargs, expectation, expected):
    with expectation:
        connection = connection_from_config(app_config, prefix="rabbitmq.", **kwargs)
    for attr, value in expected.items():
        assert getattr(connection, attr) == value


class TestKombuThriftSerializer:
    @pytest.fixture
    def serializer(self):
        return KombuThriftSerializer[ExampleStruct](ExampleStruct)

    @pytest.fixture
    def req(self):
        return ExampleStruct(string_field="foo", int_field=42)

    @pytest.fixture
    def req_bytes(self):
        return b"\x0b\x00\x01\x00\x00\x00\x03foo\n\x00\x02\x00\x00\x00\x00\x00\x00\x00*\x00"

    def test_name(self, serializer):
        assert serializer.name == "thrift-ExampleStruct"

    def test_serialize(self, serializer, req, req_bytes):
        serialized = serializer.serialize(req)
        assert isinstance(serialized, bytes)
        assert serialized == req_bytes

    def test_deserialize(self, serializer, req, req_bytes):
        request = serializer.deserialize(req_bytes)
        assert isinstance(request, ExampleStruct)
        assert request == req

    def test_serialize_errors(self, serializer):
        with pytest.raises(TypeError):
            serializer.serialize({"foo": "bar"})

    @pytest.mark.parametrize(
        "input,exc", [(None, EOFError), (b"foo", TypeError), ("foo", TypeError), (1, TypeError)]
    )
    def test_deserialize_errors(self, input, exc, serializer):
        with pytest.raises(exc):
            serializer.deserialize(input)



import unittest

from unittest import mock

import pytest

from prometheus_client import REGISTRY

from baseplate import ServerSpan
from baseplate.lib.prometheus_metrics import getHTTPSuccessLabel
from baseplate.observers.prometheus import PrometheusBaseplateObserver
from baseplate.observers.prometheus import PrometheusClientSpanObserver
from baseplate.observers.prometheus import PrometheusServerSpanObserver


class TestException(Exception):
    pass


@pytest.mark.parametrize(
    "protocol,client_or_server,observer_cls,labels",
    (
        (
            "thrift",
            "server",
            PrometheusServerSpanObserver,
            {
                "latency_labels": {"thrift_method": "", "thrift_success": "true"},
                "requests_labels": {
                    "thrift_method": "",
                    "thrift_success": "true",
                    "thrift_exception_type": "",
                    "thrift_baseplate_status": "",
                    "thrift_baseplate_status_code": "",
                },
                "active_labels": {"thrift_method": ""},
            },
        ),
        (
            "thrift",
            "client",
            PrometheusClientSpanObserver,
            {
                "latency_labels": {"thrift_slug": "", "thrift_success": "true"},
                "requests_labels": {
                    "thrift_slug": "",
                    "thrift_success": "true",
                    "thrift_exception_type": "",
                    "thrift_baseplate_status": "",
                    "thrift_baseplate_status_code": "",
                },
                "active_labels": {
                    "thrift_slug": "",
                    "thrift_method": "",
                },
            },
        ),
        (
            "http",
            "server",
            PrometheusServerSpanObserver,
            {
                "requests_labels": {
                    "http_response_code": "",
                    "http_method": "",
                    "http_endpoint": "",
                    "http_success": "false",
                },
                "latency_labels": {"http_method": "", "http_endpoint": "", "http_success": "false"},
                "active_labels": {"http_method": "", "http_endpoint": ""},
            },
        ),
        (
            "http",
            "client",
            PrometheusClientSpanObserver,
            {
                "latency_labels": {
                    "http_method": "",
                    "http_success": "false",
                    "http_slug": "",
                },
                "requests_labels": {
                    "http_method": "",
                    "http_success": "false",
                    "http_response_code": "",
                    "http_slug": "",
                },
                "active_labels": {
                    "http_method": "",
                    "http_slug": "",
                },
            },
        ),
    ),
)
def test_observer_metrics(protocol, client_or_server, observer_cls, labels):
    before_start = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_latency_seconds_count", labels.get("latency_labels", "")
    )

    assert before_start is None
    before_start = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_requests_total", labels.get("requests_labels", "")
    )
    assert before_start is None
    before_start = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_active_requests", labels.get("active_labels", "")
    )
    assert before_start is None

    observer = observer_cls()
    observer.on_set_tag("protocol", protocol)
    assert observer.get_prefix() == f"{protocol}_{client_or_server}"

    observer.on_start()
    after_start = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_latency_seconds_count", labels.get("latency_labels", "")
    )
    assert after_start is None
    after_start = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_requests_total", labels.get("requests_labels", "")
    )
    assert after_start is None
    after_start = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_active_requests", labels.get("active_labels", "")
    )
    assert after_start == 1.0

    observer.on_finish(None)
    after_done = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_latency_seconds_count", labels.get("latency_labels", "")
    )
    assert after_done == 1.0
    after_done = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_requests_total", labels.get("requests_labels", "")
    )
    assert after_done == 1.0
    after_done = REGISTRY.get_sample_value(
        f"{protocol}_{client_or_server}_active_requests", labels.get("active_labels", "")
    )
    assert after_done == 0.0


class ObserverTests(unittest.TestCase):
    def test_create_server_span(self):
        mock_server_span = mock.Mock(spec=ServerSpan)
        mock_server_span.name = "name"
        mock_context = mock.Mock()

        observer = PrometheusBaseplateObserver()
        observer.on_server_span_created(mock_context, mock_server_span)
        mock_server_span.set_tag("protocol", "thrift")

        self.assertEqual(mock_server_span.register.call_count, 1)

    def test_http_success_label(self):
        self.assertEqual("true", getHTTPSuccessLabel(222))
        self.assertEqual("true", getHTTPSuccessLabel(200))
        self.assertEqual("true", getHTTPSuccessLabel(399))
        self.assertEqual("false", getHTTPSuccessLabel(199))
        self.assertEqual("false", getHTTPSuccessLabel(400))
        self.assertEqual("false", getHTTPSuccessLabel(111))
        self.assertEqual("false", getHTTPSuccessLabel(418))



from typing import Any
from typing import Dict

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram


# default_latency_buckets creates the default bucket values for time based histogram metrics.
# we want this to match the baseplate.go default_buckets, ref: https://github.com/reddit/baseplate.go/blob/master/prometheusbp/metrics.go.
# start is the value of the lowest bucket.
# factor is amount to multiply the previous bucket by to get the value for the next bucket.
# count is the number of buckets created.
start = 0.0001
factor = 2.5
count = 14
# creates 14 buckets from 100us ~ 14.9s.
default_latency_buckets = [start * factor ** i for i in range(count)]

# Default buckets for size base histograms, from <=8 bytes to 4mB in 20
# increments (8*2^i).  Larger requests go in the +Inf bucket.
default_size_start = 8
default_size_factor = 2
default_size_count = 20
default_size_buckets = [
    default_size_start * default_size_factor ** i for i in range(default_size_count)
]

# http client labels and metrics
http_client_latency_labels = [
    "http_method",
    "http_success",
    "http_slug",
]
http_client_requests_total_labels = [
    "http_method",
    "http_success",
    "http_response_code",
    "http_slug",
]
http_client_active_requests_labels = [
    "http_method",
    "http_slug",
]

# Latency histogram of HTTP calls made by clients
# buckets are defined above (from 100µs to ~14.9s)
http_client_latency_seconds = Histogram(
    "http_client_latency_seconds",
    "Latency histogram of HTTP calls made by clients",
    http_client_latency_labels,
    buckets=default_latency_buckets,
)

# Counter counting total HTTP requests started by a given client
http_client_requests_total = Counter(
    "http_client_requests_total",
    "Total number of HTTP requests started by a given client",
    http_client_requests_total_labels,
)

# Gauge showing current number of active requests by a given client
http_client_active_requests = Gauge(
    "http_client_active_requests",
    "Number of active requests for a given client",
    http_client_active_requests_labels,
)


class PrometheusHTTPClientMetrics:
    def __init__(self) -> None:
        pass

    def latency_seconds_metric(self, tags: Dict[str, Any]) -> Histogram:
        return http_client_latency_seconds.labels(
            http_method=tags.get("http.method", ""),
            http_success=getHTTPSuccessLabel(int(tags.get("http.status_code", "0"))),
            http_slug=tags.get("http.slug", ""),
        )

    def requests_total_metric(self, tags: Dict[str, Any]) -> Counter:
        return http_client_requests_total.labels(
            http_method=tags.get("http.method", ""),
            http_success=getHTTPSuccessLabel(int(tags.get("http.status_code", "0"))),
            http_response_code=tags.get("http.status_code", ""),
            http_slug=tags.get("http.slug", ""),
        )

    def active_requests_metric(self, tags: Dict[str, Any]) -> Gauge:
        return http_client_active_requests.labels(
            http_method=tags.get("http.method", ""),
            http_slug=tags.get("http.slug", ""),
        )

    def get_latency_seconds_metric(self) -> Histogram:
        return http_client_latency_seconds

    def get_requests_total_metric(self) -> Counter:
        return http_client_requests_total

    def get_active_requests_metric(self) -> Gauge:
        return http_client_active_requests


# thrift server labels
thrift_server_latency_labels = [
    "thrift_method",
    "thrift_success",
]
thrift_server_requests_total_labels = [
    "thrift_method",
    "thrift_success",
    "thrift_exception_type",
    "thrift_baseplate_status",
    "thrift_baseplate_status_code",
]
thrift_server_active_requests_labels = ["thrift_method"]

# thrift server metrics
thrift_server_latency_seconds = Histogram(
    "thrift_server_latency_seconds",
    "RPC latencies",
    thrift_server_latency_labels,
    buckets=default_latency_buckets,
)

thrift_server_requests_total = Counter(
    "thrift_server_requests_total",
    "Total RPC request count",
    thrift_server_requests_total_labels,
)

thrift_server_active_requests = Gauge(
    "thrift_server_active_requests",
    "The number of in-flight requests being handled by the service",
    thrift_server_active_requests_labels,
)


class PrometheusThriftServerMetrics:
    def __init__(self) -> None:
        pass

    def latency_seconds_metric(self, tags: Dict[str, str]) -> Any:
        """Return the latency_seconds metrics with labels set"""
        return thrift_server_latency_seconds.labels(
            thrift_method=tags.get("thrift.method", ""),
            thrift_success=tags.get("success", ""),
        )

    def requests_total_metric(self, tags: Dict[str, str]) -> Any:
        """Return the requests_total metrics with labels set"""
        return thrift_server_requests_total.labels(
            thrift_method=tags.get("thrift.method", ""),
            thrift_success=tags.get("success", ""),
            thrift_exception_type=tags.get("exception_type", ""),
            thrift_baseplate_status=tags.get("thrift.status", ""),
            thrift_baseplate_status_code=tags.get("thrift.status_code", ""),
        )

    def active_requests_metric(self, tags: Dict[str, str]) -> Any:
        """Return the active_requests metrics with labels set"""
        return thrift_server_active_requests.labels(
            thrift_method=tags.get("thrift.method", ""),
        )

    def get_latency_seconds_metric(self) -> Histogram:
        """Return the latency_seconds metrics"""
        return thrift_server_latency_seconds

    def get_requests_total_metric(self) -> Counter:
        """Return the requests_total metrics"""
        return thrift_server_requests_total

    def get_active_requests_metric(self) -> Gauge:
        """Return the active_requests metrics"""
        return thrift_server_active_requests


thrift_client_active_gauge = Gauge(
    "thrift_client_active_requests", "Current in-flight requests", ["thrift_slug", "thrift_method"]
)

thrift_client_latency_histogram = Histogram(
    "thrift_client_latency_seconds",
    "Latency of thrift client requests",
    ["thrift_slug", "thrift_success"],
    buckets=default_latency_buckets,
)

thrift_client_requests_counter = Counter(
    "thrift_client_requests_total",
    "Total number of outgoing requests",
    [
        "thrift_slug",
        "thrift_success",
        "thrift_exception_type",
        "thrift_baseplate_status",
        "thrift_baseplate_status_code",
    ],
)


class PrometheusThriftClientMetrics:
    def __init__(self) -> None:
        pass

    def active_requests_metric(self, tags: Dict) -> Gauge:
        return thrift_client_active_gauge.labels(
            thrift_slug=tags.get("slug", ""), thrift_method=tags.get("method", "")
        )

    def requests_total_metric(self, tags: Dict) -> Counter:
        return thrift_client_requests_counter.labels(
            thrift_slug=tags.get("slug", ""),
            thrift_success=tags.get("success", ""),
            thrift_exception_type=tags.get("exception_type", ""),
            thrift_baseplate_status=tags.get("thrift_status", ""),
            thrift_baseplate_status_code=tags.get("thrift_status_code", ""),
        )

    def latency_seconds_metric(self, tags: Dict) -> Histogram:
        return thrift_client_latency_histogram.labels(
            thrift_slug=tags.get("slug", ""), thrift_success=tags.get("success", "")
        )

    def get_latency_seconds_metric(self) -> Histogram:
        """Return the latency_seconds metrics"""
        return thrift_client_latency_histogram

    def get_requests_total_metric(self) -> Counter:
        """Return the requests_total metrics"""
        return thrift_client_requests_counter

    def get_active_requests_metric(self) -> Gauge:
        """Return the active_requests metrics"""
        return thrift_client_active_gauge


# http server labels and metrics
http_server_histogram_labels = [
    "http_method",
    "http_endpoint",
    "http_success",
]

http_server_requests_total_labels = [
    "http_method",
    "http_endpoint",
    "http_success",
    "http_response_code",
]

http_server_active_requests_labels = [
    "http_method",
    "http_endpoint",
]

http_server_latency_seconds = Histogram(
    "http_server_latency_seconds",
    "Time spent processing requests",
    http_server_histogram_labels,
    buckets=default_latency_buckets,
)

http_server_request_size_bytes = Histogram(
    "http_server_request_size_bytes",
    "Size of incoming requests in bytes",
    http_server_histogram_labels,
    buckets=default_size_buckets,
)

http_server_response_size_bytes = Histogram(
    "http_server_response_size_bytes",
    "Size of outgoing responses in bytes",
    http_server_histogram_labels,
    buckets=default_size_buckets,
)

http_server_requests_total = Counter(
    "http_server_requests_total",
    "Total number of request handled",
    http_server_requests_total_labels,
)
http_server_active_requests = Gauge(
    "http_server_active_requests",
    "Current requests in flight",
    http_server_active_requests_labels,
)


class PrometheusHTTPServerMetrics:
    def __init__(self) -> None:
        pass

    def latency_seconds_metric(self, tags: Dict[str, Any]) -> Histogram:
        return http_server_latency_seconds.labels(
            http_method=tags.get("http.method", ""),
            http_endpoint=tags.get("http.route", ""),
            http_success=getHTTPSuccessLabel(int(tags.get("http.status_code", "0"))),
        )

    def requests_total_metric(self, tags: Dict[str, Any]) -> Counter:
        return http_server_requests_total.labels(
            http_method=tags.get("http.method", ""),
            http_endpoint=tags.get("http.route", ""),
            http_success=getHTTPSuccessLabel(int(tags.get("http.status_code", "0"))),
            http_response_code=tags.get("http.status_code", ""),
        )

    def active_requests_metric(self, tags: Dict[str, Any]) -> Gauge:
        return http_server_active_requests.labels(
            http_method=tags.get("http.method", ""),
            http_endpoint=tags.get("http.route", ""),
        )

    def request_size_bytes_metric(self, tags: Dict[str, Any]) -> Histogram:
        return http_server_request_size_bytes.labels(
            http_method=tags.get("http.method", ""),
            http_endpoint=tags.get("http.route", ""),
            http_success=getHTTPSuccessLabel(int(tags.get("http.status_code", "0"))),
        )

    def response_size_bytes_metric(self, tags: Dict[str, Any]) -> Histogram:
        return http_server_response_size_bytes.labels(
            http_method=tags.get("http.method", ""),
            http_endpoint=tags.get("http.route", ""),
            http_success=getHTTPSuccessLabel(int(tags.get("http.status_code", "0"))),
        )

    def get_latency_seconds_metric(self) -> Any:
        return http_server_latency_seconds

    def get_requests_total_metric(self) -> Any:
        return http_server_requests_total

    def get_active_requests_metric(self) -> Any:
        return http_server_active_requests


def getHTTPSuccessLabel(httpStatusCode: int) -> str:
    """
    The HTTP success label is "true" if the status code is 2xx or 3xx, "false" otherwise.
    """
    return str(200 <= httpStatusCode < 400).lower()



import unittest

from unittest import mock

try:
    import cassandra

    del cassandra
except ImportError:
    raise unittest.SkipTest("cassandra-driver is not installed")

import baseplate
from baseplate.lib.config import ConfigurationError
from baseplate.clients.cassandra import cluster_from_config, CassandraSessionAdapter
from baseplate.lib.secrets import SecretsStore


class ClusterFromConfigTests(unittest.TestCase):
    def test_empty_config(self):
        with self.assertRaises(ConfigurationError):
            cluster_from_config({})

    def test_contact_points(self):
        cluster = cluster_from_config({"cassandra.contact_points": "127.0.1.1, 127.0.1.2"})

        self.assertEqual(cluster.contact_points, ["127.0.1.1", "127.0.1.2"])

    def test_port(self):
        cluster = cluster_from_config(
            {"cassandra.contact_points": "127.0.1.1", "cassandra.port": "9999"}
        )

        self.assertEqual(cluster.port, 9999)

    def test_kwarg_passthrough(self):
        cluster = cluster_from_config({"cassandra.contact_points": "127.0.0.1"}, protocol_version=3)

        self.assertEqual(cluster.protocol_version, 3)

    def test_alternate_prefix(self):
        cluster = cluster_from_config(
            {"noodle.contact_points": "127.0.1.1, 127.0.1.2"}, prefix="noodle."
        )

        self.assertEqual(cluster.contact_points, ["127.0.1.1", "127.0.1.2"])

    def test_credentials_but_no_secrets(self):
        with self.assertRaises(TypeError):
            cluster_from_config(
                {
                    "cassandra.contact_points": "127.0.0.1",
                    "cassandra.credentials_secret": "secret/foo/bar",
                }
            )

    def test_credentials(self):
        secrets = mock.Mock(autospec=SecretsStore)
        cluster = cluster_from_config(
            {
                "cassandra.contact_points": "127.0.0.1",
                "cassandra.credentials_secret": "secret/foo/bar",
            },
            secrets=secrets,
        )
        self.assertIsNotNone(cluster.auth_provider)


class CassandraSessionAdapterTests(unittest.TestCase):
    def setUp(self):
        self.session = mock.MagicMock()
        self.prepared_statements = {}
        self.mock_server_span = mock.MagicMock(spec=baseplate.ServerSpan)
        self.adapter = CassandraSessionAdapter(
            "test", self.mock_server_span, self.session, self.prepared_statements
        )

    def test_prepare(self):
        statement = "SELECT foo from bar;"
        self.adapter.prepare(statement, cache=True)
        self.adapter.prepare(statement, cache=True)
        # Assert that when preparing an identical statement twice, the CQL
        # session only gets one prepare.
        self.session.prepare.assert_called_once_with(statement)
        self.adapter.prepare(statement, cache=False)
        # Assert that preparing an identical statement with cache=False always
        # prepares.
        calls = [mock.call(statement), mock.call(statement)]
        self.session.prepare.assert_has_calls(calls)



import unittest

from unittest import mock

try:
    from sqlalchemy.engine.url import URL
except ImportError:
    raise unittest.SkipTest("sqlalchemy is not installed")

from baseplate.clients.sqlalchemy import engine_from_config
from baseplate.testing.lib.secrets import FakeSecretsStore


class EngineFromConfigTests(unittest.TestCase):
    def setUp(self):
        secrets = FakeSecretsStore(
            {
                "secrets": {
                    "secret/sql/account": {
                        "type": "credential",
                        "username": "reddit",
                        "password": "password",
                    }
                }
            }
        )
        self.secrets = secrets

    def test_url(self):
        engine = engine_from_config({"database.url": "sqlite://"})
        self.assertEqual(engine.url, URL("sqlite"))

    @mock.patch("baseplate.clients.sqlalchemy.create_engine")
    def test_credentials(self, create_engine_mock):
        engine_from_config(
            {
                "database.url": "postgresql://fizz:buzz@localhost:9000/db",
                "database.credentials_secret": "secret/sql/account",
                "database.pool_recycle": "60",
                "database.pool_size": "10",
                "database.max_overflow": "5",
            },
            self.secrets,
        )
        create_engine_mock.assert_called_once_with(
            URL(
                drivername="postgresql",
                username="reddit",
                password="password",
                host="localhost",
                port="9000",
                database="db",
            ),
            pool_recycle=60,
            pool_size=10,
            max_overflow=5,
        )

    @mock.patch("baseplate.clients.sqlalchemy.create_engine")
    def test_credentials_no_secrets(self, create_engine_mock):
        with self.assertRaises(TypeError):
            engine_from_config(
                {
                    "database.url": "postgresql://fizz:buzz@localhost:9000/db",
                    "database.credentials_secret": "secret/sql/account",
                }
            )
        self.assertEqual(create_engine_mock.call_count, 0)



import unittest

try:
    import rediscluster
except ImportError:
    raise unittest.SkipTest("redis-py-cluster is not installed")

from baseplate.lib.config import ConfigurationError
from baseplate.clients.redis_cluster import cluster_pool_from_config

from baseplate.clients.redis_cluster import ClusterRedisClient
from baseplate import Baseplate
from . import TestBaseplateObserver, get_endpoint_or_skip_container

redis_endpoint = get_endpoint_or_skip_container("redis-cluster-node", 7000)


# This belongs on the unit tests section but the client class attempts to initialise
# the list of nodes when being instantiated so it's simpler to test here with a redis
# cluster available
class ClusterPoolFromConfigTests(unittest.TestCase):
    def test_empty_config(self):
        with self.assertRaises(ConfigurationError):
            cluster_pool_from_config({})

    def test_basic_url(self):
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})

        self.assertEqual(pool.nodes.startup_nodes[0]["host"], "redis-cluster-node")
        self.assertEqual(pool.nodes.startup_nodes[0]["port"], "7000")

    def test_timeouts(self):
        pool = cluster_pool_from_config(
            {
                "rediscluster.url": f"redis://{redis_endpoint}/0",
                "rediscluster.timeout": "30 seconds",
            }
        )

        self.assertEqual(pool.timeout, 30)

    def test_max_connections(self):
        pool = cluster_pool_from_config(
            {
                "rediscluster.url": f"redis://{redis_endpoint}/0",
                "rediscluster.max_connections": "300",
            }
        )

        self.assertEqual(pool.max_connections, 300)

    def test_max_connections_default(self):
        # https://github.com/Grokzen/redis-py-cluster/issues/435
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})

        self.assertEqual(pool.max_connections, 50)

    def test_kwargs_passthrough(self):
        pool = cluster_pool_from_config(
            {"rediscluster.url": f"redis://{redis_endpoint}/0"}, example="present"
        )

        self.assertEqual(pool.connection_kwargs["example"], "present")

    def test_alternate_prefix(self):
        pool = cluster_pool_from_config(
            {"noodle.url": f"redis://{redis_endpoint}/0"}, prefix="noodle."
        )
        self.assertEqual(pool.nodes.startup_nodes[0]["host"], "redis-cluster-node")
        self.assertEqual(pool.nodes.startup_nodes[0]["port"], "7000")

    def test_only_primary_available(self):
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})
        node_list = [pool.get_node_by_slot(slot=1, read_command=False) for _ in range(0, 100)]

        # The primary is on port 7000 so that's the only port we expect to see
        self.assertTrue(all(node["port"] == 7000 for node in node_list))

    def test_read_from_replicas(self):
        pool = cluster_pool_from_config({"rediscluster.url": f"redis://{redis_endpoint}/0"})

        node_list = [pool.get_node_by_slot(slot=1, read_command=True) for _ in range(0, 100)]

        # Both replicas and primary are available, so we expect to see some non-primaries here
        self.assertTrue(any(node["port"] != 7000 for node in node_list))


class RedisClusterIntegrationTests(unittest.TestCase):
    def setUp(self):
        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate(
            {
                "rediscluster.url": f"redis://{redis_endpoint}/0",
                "rediscluster.timeout": "1 second",
                "rediscluster.max_connections": "4",
            }
        )
        baseplate.register(self.baseplate_observer)
        baseplate.configure_context({"rediscluster": ClusterRedisClient()})

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_simple_command(self):
        with self.server_span:
            result = self.context.rediscluster.ping()

        self.assertTrue(result)

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "rediscluster.PING")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)

    def test_error(self):
        with self.server_span:
            with self.assertRaises(rediscluster.RedisClusterException):
                self.context.rediscluster.execute_command("crazycommand")

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNotNone(span_observer.on_finish_exc_info)

    def test_lock(self):
        with self.server_span:
            with self.context.rediscluster.lock("foo-lock"):
                pass

        server_span_observer = self.baseplate_observer.get_only_child()

        self.assertGreater(len(server_span_observer.children), 0)
        for span_observer in server_span_observer.children:
            self.assertTrue(span_observer.on_start_called)
            self.assertTrue(span_observer.on_finish_called)

    def test_pipeline(self):
        with self.server_span:
            with self.context.rediscluster.pipeline("foo") as pipeline:
                pipeline.set("foo", "bar")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.get("foo")
                pipeline.delete("foo")
                pipeline.execute()

        server_span_observer = self.baseplate_observer.get_only_child()
        span_observer = server_span_observer.get_only_child()
        self.assertEqual(span_observer.span.name, "rediscluster.pipeline_foo")
        self.assertTrue(span_observer.on_start_called)
        self.assertTrue(span_observer.on_finish_called)
        self.assertIsNone(span_observer.on_finish_exc_info)
